

import numpy as np
import time
import tqdm
import os

from photonpy import Context, Gaussian, Estimator, Dataset
import photonpy.cpp.spotdetect as spotdetect
from photonpy.cpp.calib import GainOffset_Calib
from photonpy.cpp.calib import GainOffsetImage_Calib
from photonpy.cpp.estim_queue import EstimQueue
from photonpy.cpp.roi_queue import ROIQueue
from photonpy.cpp.image_proc import ROIExtractor
import photonpy.utils.multipart_tiff as tiff
import tifffile
import matplotlib.pyplot as plt

from scipy.interpolate import InterpolatedUnivariateSpline

def end_of_file(f):
    curpos = f.tell()
    f.seek(0,2)
    file_size = f.tell()
    f.seek(curpos,0)
    return curpos == file_size 


def detect_spots(sdcfg, calib, movie, sumframes, output_fn, batch_size, ctx:Context, numThreads=3):
    sm = spotdetect.SpotDetectionMethods(ctx)

    with Context(ctx.smlm) as lq_ctx, open(output_fn, "wb") as f:
        roishape = [sdcfg.roisize,sdcfg.roisize]
        
        numframes = 0
        numrois = 0
        
        for i,img in enumerate(movie):
            if i==0:
                q,rq = sm.CreateQueue(img.shape, roishape, sdcfg, calib=calib,sumframes=sumframes, ctx=lq_ctx, numThreads=numThreads)
        
            def save_rois(rois_info, pixels):
                np.save(f, rois_info, allow_pickle=False)
                np.save(f, pixels, allow_pickle=False)
                nonlocal numrois
                numrois += len(rois_info)

            q.PushFrame(img)
            numframes += 1
            
            rl = rq.Length()
            if rl>batch_size:
                save_rois(*rq.Fetch())
                   
        while q.NumFinishedFrames() < numframes//sumframes:
            time.sleep(0.1)
        
        if rq.Length()>0:
            save_rois(*rq.Fetch())
            
        return numrois, numframes


def load_rois_iterator(rois_fn):
    """
    Load rois sequentially so we can deal with very large datasets
    """
    with open(rois_fn, "rb") as f:
        while not end_of_file(f):
            rois_info = np.load(f)
            pixels = np.load(f)
            yield rois_info, pixels
            
    
def load_rois(rois_fn):
    rois_info = []
    pixels = []
    for ri,px in load_rois_iterator(rois_fn):
        rois_info.append(ri)
        pixels.append(px)
    
    return np.concatenate(rois_info), np.concatenate(pixels)


def extract_rois_iterator(movie, roipos, roiframe, calib, roisize, 
                          ctx:Context, minBatchSize):
    """
    Extract predefined ROIs from a TIFF or other image iterator. 
    Calib: camera calibration as created with create_calib_obj
    roipos: Corner positions [[y,x]...]
    TODO: Save to disk in batches to remove memory bottleneck
    """
    roilist = np.zeros(len(roipos),dtype=ROIExtractor.ROIType)
    roilist['cornerpos'] = roipos
    roilist['startframe'] = roiframe
    roilist['numframes'] = 1

    with Context(ctx.smlm) as ex_ctx:
        q = None        
        numframes = 0
        for i,img in enumerate(movie):
            if i == 0:
                q = ROIExtractor(img.shape, roilist, 1, roisize, calib, ex_ctx)

            q.PushFrame(img)
            numframes += 1
 
            resultcount = q.GetResultCount()
            if resultcount > minBatchSize:
                rois,frames = q.GetResults(resultcount)
                yield rois,frames

        while not q.IsIdle():
            time.sleep(0.1)

            resultcount = q.GetResultCount()
            if resultcount > minBatchSize:
                rois,frames = q.GetResults(resultcount)
                yield rois,frames

        resultcount = q.GetResultCount()
        if resultcount>0:
            rois,frames = q.GetResults(resultcount)
            yield rois,frames


def localize_rois(rois_data_source, psf, initial_estim=None, constants=None):
    """
    Runs PSF centroid estimation on the given ROIs using the PSF.
    The PSF must be created with cuda=True
    rois_data_source can either be a file generated by detect_rois, or 
    a tuple with the ROI data (rois_info, pixels)
    """
    
    if type(rois_data_source) == str:
        iterator = load_rois_iterator(rois_data_source)
    else:
        iterator = [rois_data_source]

    framenums = []    
    with EstimQueue(psf,batchSize=1024) as queue:
        count = 0
        for rois_info, pixels in iterator:
            roipos = np.zeros((len(rois_info),2))
            roipos[:,0] = rois_info['y']
            roipos[:,1] = rois_info['x']
            
            framenum = rois_info['id']
            framenums.append(framenum)
            if initial_estim is not None:
                initial = initial_estim[count:count+len(pixels)]
            else:
                initial = None
                
            if constants is not None:
                batch_constants = constants[count:count+len(pixels)]
            else:
                batch_constants = None
                
            queue.Schedule(pixels, ids=np.arange(count,count+len(pixels)), roipos=roipos, initial=initial, constants=batch_constants)
            count += len(pixels)
            
        queue.WaitUntilDone()
        r = queue.GetResults(getSampleData=True)
        r.SortByID(isUnique=True) # reorder back to original ROI order
        r.ids = np.concatenate(framenums)

    return r

def create_calib_obj(gain,offset,imgshape,ctx):        
    if type(gain)==str:
        print(f'estimating gain from light {gain} and dark {offset} frames')
        light = tifffile.imread(gain)
        offset = tifffile.imread(offset)
        
        if not np.array_equal(imgshape, light.shape[1:]):
            raise ValueError(f'Camera light frames calibration data ({light.shape[1:]}) does not match with expected image size ({imgshape})')
            
        if not np.array_equal(imgshape, offset.shape[1:]):
            raise ValueError(f'Camera offset calibration ({offset.shape[1:]}) does not match with expected image size ({imgshape})')

        offset = np.mean(offset,0)
        sig = light-offset
        v = np.var(sig, 0)
        m = np.mean(sig,0)
        
        gain = v/m
        gain[gain==0] = np.mean(gain)
        print(f'mean camera gain: {np.mean(gain):.2f} ADU/photons offset: {np.mean(offset):.2f}',flush=True)
        

    if type(offset)==str:
        print(f'using mean values from {offset} as camera offset',flush=True)
        offset=tiff.get_tiff_mean(offset)

        if type(gain)!=str:
            gain = np.ones(imgshape)*gain
    
    if( type(offset)==np.ndarray):
        calib = GainOffsetImage_Calib(gain, offset, ctx)
    else:
        calib = GainOffset_Calib(gain, offset, ctx) 
    
    return calib

def _summed_movie(movie, sumframes):
    img = None
    f = 0
    for m in movie:
        if f == 0:
            img = m.copy()
        else:
            img += m

        f += 1
        if f == sumframes:
            img_ = img
            img = None
            f = 0
            yield img_
            
    


class Localizer2D:
    """
    Perform localization on a tiff with a 2D Gaussian PSF model
    """    
    def __init__(self):
        ...
        
    def process(self, fn, cfg, output_file=None, progress_cb=None):
        rois_output_fn = os.path.splitext(output_file)[0]+"-rois.npy"
        
        self.cfg=cfg
        
        roisize = cfg['roisize']
        threshold = cfg['threshold']
        gain = cfg['gain']
        offset = cfg['offset']
        startframe = cfg['startframe'] if 'startframe' in cfg else 0
        maxframes = cfg['maxframes'] if 'maxframes' in cfg else -1
        sumframes = cfg['sumframes'] if 'sumframes' in  cfg else 1
        maxChiSquare = cfg['maxchisq'] if 'maxchisq' in cfg else None
        sigmaFramesPerBin = cfg['sigmaframesperbin'] if 'sigmaframesperbin' in cfg else None
        spotDetectSigma = cfg['spotdetectsigma']
        
        fovOffsets = cfg['fov_offsets'] if 'fov_offsets' in cfg else None
        useTiltedBg = cfg['tilted_bg'] if 'tilted_bg' in cfg else None
        
        def progreport(msg):
            if progress_cb is not None:
                return progress_cb(msg,0)
            return True
        
        with Context() as ctx:
            imgshape = tiff.tiff_get_image_size(fn)
            
            print(imgshape)
    
            gaussian = Gaussian(ctx)
                
            movie = tiff.tiff_read_file(fn, startframe, maxframes, progress_cb)
            if sumframes > 1:
                movie = _summed_movie(movie, sumframes)
                offset *= sumframes

            spotDetector = spotdetect.SpotDetector(spotDetectSigma, roisize, threshold)
            calib = create_calib_obj(gain,offset,imgshape,ctx)
            
            numrois,_ = detect_spots(spotDetector, calib, movie, 1, rois_output_fn, batch_size=20000, ctx=ctx)
            
            if numrois == 0:
                raise ValueError('No spots found')
    
            if not progreport(f'Fitting 2D Gaussian to {numrois} ROIs (including PSF sigma) 1/2'):
                return None
            
            psf = gaussian.CreatePSF_XYIBg(roisize, spotDetectSigma, True)
            qr = localize_rois(rois_output_fn, psf)
            framenum = qr.ids

            if not progreport(f'Fitting 2D Gaussian to {numrois} ROIs (including PSF sigma) 2/2'):
                return None
    
            # Re-estimate including sigma (x,y) fits
            estim = np.zeros((len(qr.estim), 6))
            estim[:,:4] = qr.estim
            estim[:,4:] = spotDetectSigma
            psf_sigma = gaussian.CreatePSF_XYIBgSigmaXY(roisize, spotDetectSigma, True)
            qr_sigma = localize_rois(rois_output_fn, psf_sigma, initial_estim=estim)
    
            # Estimate per-frame sigma and interpolate using splines to account for Z drift 
            ds = Dataset.fromQueueResults(qr_sigma, imgshape)
            self.ds_sigma_fits = ds
            
            chiSqThreshold = maxChiSquare*roisize**2
            ds = ds[ds.data.chisq<chiSqThreshold]

            if len(ds) == 0:
                raise ValueError('PSF Sigma fits failed')
                            
            numframes = np.max(framenum)+1
            ds.data.frame = np.maximum((ds.data.frame / sigmaFramesPerBin - 0.5).astype(np.int32),0)
            frames = ds.indicesPerFrame()
            self.medianSigma = np.array([np.median(ds.data.estim.sigma[idx],0) for idx in frames])
                
            self.sigma_t = (0.5+np.arange(len(frames))) * sigmaFramesPerBin
            spl_x = InterpolatedUnivariateSpline(self.sigma_t, self.medianSigma[:,0], k=2)
            spl_y = InterpolatedUnivariateSpline(self.sigma_t, self.medianSigma[:,1], k=2)
            
            self.sigma = np.zeros((numframes,2))
            self.sigma[:,0] = spl_x(np.arange(numframes))
            self.sigma[:,1] = spl_y(np.arange(numframes))
                                
            # Re-estimate using known sigmas
            if not progreport('Fitting 2D Gaussian to ROIs (fixed sigma)'):
                return None
            psf = gaussian.CreatePSF_XYIBg(roisize, sigma=None, cuda=True)
            roi_sigmas = self.sigma[framenum]
            r = localize_rois(rois_output_fn, psf, constants=roi_sigmas)
                    
            if maxChiSquare is not None:
                print(f"Filtering on chi-square at threshold {maxChiSquare}*{roisize*roisize}...")
                self.chiSqThreshold = maxChiSquare*roisize**2
                self.unfilteredChisq = r.chisq*1
                r.Filter(r.chisq<self.chiSqThreshold)
                
            print(f"Filtering {len(r.estim)} spots at ROI boundary and iterations")
            minX = 2.1
            minY = 2.1
            r.FilterXY(minX,minY,roisize-minX-1, roisize-minY-1)
            r.Filter(r.iterations<50)
                   
            nframes = np.max(r.ids)+1 if len(r.ids)>0 else 1
            print(f"Num spots: {len(r.estim)}. {len(r.estim) / nframes} spots/frame.")
            
            ds = Dataset.fromQueueResults(r, imgshape, config=cfg, imagefile=fn, sigma=self.sigma)
            if output_file is not None:
                ds.save(output_file)

            self.result=ds
            ds.config['locs_path'] = output_file
            return ds
        
    def plotSigmaFits(self):
        plt.figure()
        plt.hist(self.data.estim.sigma.T, bins=100,label=['Sigma X', 'Sigma Y'])
        plt.legend()
        
    def plotChiSquare(self):
        plt.figure()
        plt.hist(self.unfilteredChisq, bins=100, range=[0,self.chiSqThreshold*1.5], label='Chi-square values')
        plt.axvline(self.chiSqThreshold, label='Threshold')
        plt.legend()
        plt.xlabel('Chi-square value (sum((data-model)^2/model)')

    def plotSigmaTimeSeries(self):
        sigmaFramesPerBin = self.cfg['sigmaframesperbin']
        plt.figure()
        plt.plot(self.sigma_t, self.medianSigma[:,0],'o')
        plt.plot(self.sigma_t, self.medianSigma[:,1],'o')
        plt.plot(self.sigma[:,0], label='Sigma X')
        plt.plot(self.sigma[:,1], label='Sigma Y')
        plt.xlabel('Frames')
        plt.ylabel('Gaussian PSF Sigma [pixels]')
        plt.legend()
        plt.title(f'PSF Sigma vs time (using {sigmaFramesPerBin} frames per bin)')
